@@ -2,7 +2,7 @@
 /*
  * Ruby Cairo Binding
  *
- * Copyright 2005-2015 Kouhei Sutou <kou@cozmixng.org>
+ * Copyright 2005-2017 Kouhei Sutou <kou@cozmixng.org>
  * Copyright 2014 Patrick Hanevold <patrick.hanevold@gmail.com>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -465,6 +465,18 @@ cr_surface_destroy (VALUE self)
   return self;
 }
 
+static VALUE
+cr_surface_get_reference_count (VALUE self)
+{
+  cairo_surface_t *surface;
+  unsigned int reference_count;
+
+  surface = _SELF;
+  reference_count = cairo_surface_get_reference_count (surface);
+
+  return UINT2NUM (reference_count);
+}
+
 static VALUE
 cr_surface_finish (VALUE self)
 {
@@ -1177,6 +1189,111 @@ cr_pdf_surface_restrict_to_version (VALUE self, VALUE version)
   return Qnil;
 }
 #  endif
+
+#  if CAIRO_CHECK_VERSION(1, 15, 4)
+static VALUE
+cr_pdf_surface_add_outline (VALUE self,
+                            VALUE rb_parent_id,
+                            VALUE rb_name,
+                            VALUE rb_destination,
+                            VALUE rb_flags)
+{
+  cairo_surface_t *surface;
+  int parent_id;
+  const char *name;
+  const char *destination;
+  cairo_pdf_outline_flags_t flags;
+  int id;
+
+  surface = _SELF;
+  if (NIL_P (rb_parent_id))
+    parent_id = 0;
+  else
+    parent_id = NUM2INT (rb_parent_id);
+  name = RVAL2CSTR (rb_name);
+  destination = RVAL2CSTR (rb_destination);
+  if (NIL_P (rb_flags))
+    flags = 0;
+  else
+    flags = RVAL2CRPDFOUTLINEFLAGS (rb_flags);
+  id = cairo_pdf_surface_add_outline (surface,
+                                      parent_id,
+                                      name,
+                                      destination,
+                                      flags);
+  rb_cairo_surface_check_status (surface);
+
+  if (id == 0)
+    return Qnil;
+  else
+    return INT2NUM (id);
+}
+
+static VALUE
+cr_pdf_surface_set_metadata (VALUE self,
+                             VALUE rb_metadata,
+                             VALUE rb_value)
+{
+  cairo_surface_t *surface;
+  cairo_pdf_metadata_t metadata;
+  const char *value;
+
+  surface = _SELF;
+  metadata = RVAL2CRPDFMETADATA (rb_metadata);
+  switch (metadata)
+    {
+    case CAIRO_PDF_METADATA_CREATE_DATE:
+    case CAIRO_PDF_METADATA_MOD_DATE:
+      if (rb_cairo__is_kind_of (rb_value, rb_cTime))
+        {
+          ID id_iso8601;
+          CONST_ID(id_iso8601, "iso8601");
+          rb_value = rb_funcall (rb_value, id_iso8601, 0);
+        }
+      break;
+    default:
+      break;
+    }
+  value = RVAL2CSTR (rb_value);
+  cairo_pdf_surface_set_metadata (surface, metadata, value);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+
+static VALUE
+cr_pdf_surface_set_page_label (VALUE self,
+                               VALUE rb_label)
+{
+  cairo_surface_t *surface;
+  const char *label;
+
+  surface = _SELF;
+  label = RVAL2CSTR (rb_label);
+  cairo_pdf_surface_set_page_label (surface, label);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+
+static VALUE
+cr_pdf_surface_set_thumbnail_size (VALUE self,
+                                   VALUE rb_width,
+                                   VALUE rb_height)
+{
+  cairo_surface_t *surface;
+  int width;
+  int height;
+
+  surface = _SELF;
+  width = NUM2INT (rb_width);
+  height = NUM2INT (rb_height);
+  cairo_pdf_surface_set_thumbnail_size (surface, width, height);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+#  endif
 #endif
 
 #ifdef CAIRO_HAS_PS_SURFACE
@@ -1893,6 +2010,8 @@ Init_cairo_surface (void)
                     cr_surface_get_device, 0);
 #endif
   rb_define_method (rb_cCairo_Surface, "destroy", cr_surface_destroy, 0);
+  rb_define_method (rb_cCairo_Surface, "reference_count",
+                    cr_surface_get_reference_count, 0);
   rb_define_method (rb_cCairo_Surface, "finish", cr_surface_finish, 0);
   rb_define_method (rb_cCairo_Surface, "content", cr_surface_get_content, 0);
 
@@ -1980,6 +2099,26 @@ Init_cairo_surface (void)
                     cr_pdf_surface_restrict_to_version, 1);
 #  endif
 
+#  if CAIRO_CHECK_VERSION(1, 15, 4)
+  {
+    VALUE rb_mCairo_PDFOutline;
+
+    rb_mCairo_PDFOutline =
+      rb_define_module_under (rb_mCairo, "PDFOutline");
+    rb_define_const (rb_mCairo_PDFOutline, "ROOT",
+                     INT2NUM (CAIRO_PDF_OUTLINE_ROOT));
+
+    rb_define_method (rb_cCairo_PDFSurface, "add_outline",
+                      cr_pdf_surface_add_outline, 4);
+    rb_define_method (rb_cCairo_PDFSurface, "set_metadata",
+                      cr_pdf_surface_set_metadata, 2);
+    rb_define_method (rb_cCairo_PDFSurface, "set_page_label",
+                      cr_pdf_surface_set_page_label, 1);
+    rb_define_method (rb_cCairo_PDFSurface, "set_thumbnail_size",
+                      cr_pdf_surface_set_thumbnail_size, 2);
+  }
+#  endif
+
   RB_CAIRO_DEF_SETTERS (rb_cCairo_PDFSurface);
 #endif
 
